### Remote gclean execution
import os
import sys
sys.path.insert(0, os.path.abspath('../'))
##>>>import casaremote

# Remote functions needed
# finalize, update(), reset

# TODO: How to handle dictionaries? """?
def kwarg_string(*args, **kwargs):
    ret = []
    for arg in args:
        if isinstance(arg, str):
            ret.append(f"'{arg}'")
        else:
            ret.append(f"{arg}")
    for key, value in kwargs.items():
        if isinstance(value, str):
            value = f"'{value}'"
        ret.append(f"{key}={value}")
    return ", ".join(ret)

class gclean_local:

    # def _tclean( self, *args, **kwargs ):
    #     return

    # def _deconvolve( self, *args, **kwargs ):
    #     return
    #     # return ( *args, **kwargs )

    def cmds( self ):
        cmd = casaremote.create_cmd_json("self.gclean_cmds")
        gclean_id = casaremote.exe_remote_subproc(self.kc, cmd)
        ret = casaremote.get_subproc_return_val(self.kc, gclean_id)
        return ret

    def __init__(self, kc, *args, **kwargs):
        self.kc = kc
        print("In remote clean init \n")
        cmd = "t._clean = gclean(" + kwarg_string(*args, **kwargs) + ")"
        casaremote.exe_and_print(kc, cmd)

    def __next__( self ):
        cmd = casaremote.create_cmd_json("self.gclean_next")
        gclean_id = casaremote.exe_remote_subproc(self.kc, cmd)
        ret = casaremote.get_subproc_return_val(self.kc, gclean_id)
        return ret

    def restore(self):
        cmd = casaremote.create_cmd_json("self.gclean_restore")
        gclean_id = casaremote.exe_remote_subproc(self.kc, cmd)
        ret = casaremote.get_subproc_return_val(self.kc, gclean_id)
        return ret

    def has_next(self):
        cmd = casaremote.create_cmd_json("self.gclean_has_next")
        gclean_id = casaremote.exe_remote_subproc(self.kc, cmd)
        ret = casaremote.get_subproc_return_val(self.kc, gclean_id)
        return ret



import inspect
import importlib
import sys
import types

def ImportProtectedModule( name, file_mapping ):

    class module_import(types.ModuleType):
        """Import objects, functions and classes for export while avoiding requiring user to
           provide the sub-package name for the file contining the imported object, for example
           allowing:
                         from cubevis import iclean

           instead of:
                         from cubevis.app._interactive_clean import iclean

           while also not importing '_interactive_clean' (inside of __init__.py) regardless of
           whether the user actually import 'iclean' or not. This is done to allow for some
           dependencies to not be available but for the imports to still work as long as
           the dependencies which are needed for the imported classes are available.
        """

        _module = name
        _mapping = file_mapping

        def __getattr__(self, name):
            if name in self._mapping:
                m = importlib.import_module(self._mapping[name], self._module)  # import submodule
                o = getattr(m, name)                                            # find the required member
                setattr(sys.modules[self._module], name, o)                     # bind it into the package
                return o
            else:
                raise AttributeError(f'module {__name__} has no attribute {name}')

    return module_import



from cubevis.utils import is_colab
import websockets
import http
from websockets.server import serve
from websockets.http11 import Response

log_path = "/content/package_debug.txt"

def colab_cors_handler(request):
    """Answers the priming fetch once Google Proxy lets it through."""
    # LOG EVERY ATTEMPT - If this doesn't show up, traffic isn't reaching Python
    with open(log_path, "a") as f:
        f.write(f"V15 HANDLER: {request.method} to {request.path}\n")
    if "upgrade" not in request.headers.get("Connection", "").lower():
        origin = request.headers.get("Origin", "*")
        headers = {
            "Access-Control-Allow-Origin": origin,
            "Access-Control-Allow-Credentials": "true",
            "Access-Control-Allow-Methods": "GET, OPTIONS",
        }
        return Response(http.HTTPStatus.OK, "OK", headers)
    return None

def create_ws_server(callback, ip_address, port):
    if is_colab():
        print(f"Websocket startup: 0.0.0.0:{port} (CORS Enabled)")
        return serve(
            callback,
            "0.0.0.0",
            port,
            process_request=colab_cors_handler # Add this to fix the fetch error
        )
    else:
        return serve(callback, ip_address, port)



from ._tip_button import TipButton
from ._tip import Tip
from ._edit_span import EditSpan
from ._ev_text_input import EvTextInput
from ._showable import Showable
from ._shared_dict import SharedDict
from ._bokeh_app_context import BokehAppContext



### Useful for debugging
#from cubevis.bokeh import set_cubevis_lib
#set_cubevis_lib( bokeh_tables="{JSLIB}/bokeh-tables-3.6.1.js" )

import os
import ssl
import certifi
import urllib
import tarfile

from cubevis import iclean

##
## demo measurement set to use
##
ms_path = 'refim_twopoints_twochan.ms'
##
## where to fetch the demo measurement set
##
ms_url = "https://casa.nrao.edu/download/devel/casavis/data/refim_twopoints_twochan-ms.tar.gz"
##
## output image file name
##
img = 'test'

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall( )
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

if not os.path.isdir(ms_path):
    raise  RuntimeError("Failed to fetch measurement set")

print("Result = " + str( repr( iclean( vis=ms_path, imagename=img,
                                       imsize=100,
                                       cell='8.0arcsec',
                                       phasecenter="J2000 19:59:28.500 +40.44.01.50",
                                       outlierfile='test_outlier.txt',
                                       niter=50,
                                       cycleniter=10,
                                       deconvolver='hogbom',
                                       specmode='mfs',
                                       spw='0:0') ) ) )



import os
import sys
import logging

logger = logging.getLogger(__name__)

#def is_notebook() -> bool:
#    try:
#        shell = get_ipython().__class__.__name__
#        if shell == 'ZMQInteractiveShell':
#            return True   # Jupyter notebook or qtconsole
#        elif shell == 'TerminalInteractiveShell':
#            return False  # Terminal running IPython
#        else:
#            if get_ipython().__class__.__module__ == 'google.colab._shell':
#                return True   # Google Colab
#            else:
#              return False  # Other type (?)
#    except NameError:
#        return False

def is_colab():
    """
    Detects if the current environment is Google Colab.
    Combines module check (process-level) and env-var check (VM-level).
    """
    # 1. Check for the Colab-specific environment variable (VM level)
    # 2. Check for the google.colab module (Python process level)
    return os.getenv("COLAB_RELEASE_TAG") is not None or 'google.colab' in sys.modules

def is_interactive_jupyter( ) -> bool:
    """
    Detect if running in an interactive Jupyter notebook with frontend connection.

    This function distinguishes between:
    - Interactive Jupyter notebook/lab with frontend (returns True)
    - Standalone Jupyter kernel without frontend (returns False)
    - Regular Python interpreter (returns False)
    - Support Google Colab and legacy Jupyter versions.

    Returns:
        bool: True if running in interactive Jupyter with frontend, False otherwise
    """
    try:
        from IPython import get_ipython
        ipython = get_ipython()

        if ipython is None:
            logger.debug(f"\tis_interactive_jupyter<1>: False")
            return False

        # Allow Colab's 'Shell' class in addition to standard 'ZMQInteractiveShell'
        shell_class = ipython.__class__.__name__
        is_colab = 'google.colab' in sys.modules

        # Check if we're in a ZMQ-based shell (kernel) or Colab's 'Shell'
        if shell_class not in ['ZMQInteractiveShell', 'Shell'] and not is_colab:
            logger.debug(f"\tis_interactive_jupyter<2>: False")
            return False

        # If we are in Colab, the existence of the shell itself implies
        # an interactive frontend, as Colab sessions are inherently interactive.
        if is_colab:
            return True

        # Check for active frontend connection (Standard Jupyter/Legacy)
        if hasattr(ipython, 'kernel') and ipython.kernel is not None:
            kernel = ipython.kernel

            # Method 1: Check if there are active connections
            if hasattr(kernel, 'shell_socket') and kernel.shell_socket is not None:
                # For newer Jupyter versions, check connection count
                if hasattr(kernel, 'connection_count'):
                    logger.debug(f"\tis_interactive_jupyter<3>: {kernel.connection_count > 0}")
                    if kernel.connection_count > 0:
                        return True

                # For older versions, check if socket is connected
                try:
                    # Try to get socket state - if it fails, likely no frontend
                    logger.debug(f"\tis_interactive_jupyter<4>: {not socket_state}")
                    if not kernel.shell_socket.closed:
                        return True
                except AttributeError:
                    pass

            # Method 2: Check parent message (indicates interactive execution)
            if hasattr(kernel, 'get_parent') and callable(kernel.get_parent):
                try:
                    parent = kernel.get_parent()
                    # If there's a parent message, we're likely in interactive mode
                    logger.debug(f"\tis_interactive_jupyter<5>: {parent is not None and len(parent) > 0}")
                    if parent is not None and len(parent) > 0:
                        return True
                except Exception:
                    pass

            # Method 3: Check for execution context
            if hasattr(kernel, '_parent_ident') and kernel._parent_ident:
                logger.debug(f"\tis_interactive_jupyter<6>: True")
                return True

        # Fallback: Check for environment indicators (Legacy/Containerized Jupyter)
        jupyter_indicators = [
            'JPY_PARENT_PID',       # JupyterLab/Notebook sets this
            'JUPYTER_RUNTIME_DIR',
            'COLAB_RELEASE_DEVICE', # Additional Colab-specific env check
            'COLAB_GPU'             # Additional Colab-specific env check
        ]

        for indicator in jupyter_indicators:
            if indicator in os.environ:
                # Additional check: see if we can import notebook-specific modules
                try:
                    import IPython.display
                    # If we can import display and have env indicators, likely interactive
                    logger.debug(f"\tis_interactive_jupyter<7>: True")
                    return True
                except ImportError:
                    pass

        logger.debug(f"\tis_interactive_jupyter<8>: False")
        return False

    except (ImportError, Exception):
        logger.debug(f"\tis_interactive_jupyter<9>: False")
        return False



import time
import asyncio
import websockets

message_count = 0
async def timeserver( websocket ):
    global message_count
    async for message in websocket:
        message_count += 1
        print(f'{message_count}\t{message}')
        await websocket.send(time.strftime('%Y-%m-%d.%H:%M:%S'))

async def main( ):
    async with websockets.serve( timeserver, "localhost", 8765 ):
        await asyncio.Future( )

asyncio.run(main( ))



import os
import ssl
import certifi
import urllib
import tarfile

import numpy as np
from bokeh.models import CustomJS
from bokeh.plotting import figure, show
from casatools import ms as mstool
from cubevis.utils import serialize, deserialize, find_ws_address

ms_path = 'tl018b.ms'
ms_url = 'https://casa.nrao.edu/download/devel/casavis/data/tl018b_ms.tar.gz'

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall()
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

ms1 = mstool( )
ms1.open('tl018b.ms')
ms1.msselect({'scan': "91",'baseline':"OV,PT"})

data1 = ms1.getdata(['DATA'])
print( f'''                                Rows in the full MS: {ms1.nrow( )}''' )
print( f'''                Rows without polarization selection: {ms1.nrow(True)}''' )
print( f'''          Data shape without polarization selection: {data1['data'].shape}''' )
print( f'''Type of data element without polarization selection: {type(data1['data'][0][0][0])}''' )

def abort_handler( self, err ):
    print(f'''Abort signaled from browser:\n    {err}''')
    ### signal shutdown here

p = figure( title="Amplitude Plot", x_axis_label="Spectral Index", y_axis_label="Amplitude" )
print( f'''Plot size:\t {data1['data'][0].size} (x2)''' )
p.scatter( list(range(data1['data'][0].size)), np.abs(data1['data'][0].flatten( )), color="blue" )
p.scatter( list(range(data1['data'][1].size)), np.abs(data1['data'][1].flatten( )), color="red" )

show(p)



########################################################################3
#  _gclean.py
#
# Copyright (C) 2021,2022,2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
import os
import asyncio
from functools import reduce
import copy
import numpy as np
import shutil
import time
import subprocess

from casatasks.private.imagerhelpers.imager_return_dict import ImagingDict
from casatasks import deconvolve, tclean, imstat
from casatasks import casalog

###
### import check versions
###
_GCV001 = True
_GCV002 = True
_GCV003 = True
_GCV004 = True


# from casatasks.private.imagerhelpers._gclean import gclean
class gclean:
    '''gclean(...) creates a stream of convergence records which indicate
    the convergence quaility of the tclean process. The initial record
    describes the initial dirty image.
    It is designed for use with the interactive clean GUI, but it could
    be used independently. It can be used as a regular generator:
          for rec in gclean( vis='refim_point_withline.ms', imagename='test', imsize=512, cell='12.0arcsec',
                             specmode='cube', interpolation='nearest', nchan=5, start='1.0GHz', width='0.2GHz',
                             pblimit=-1e-05, deconvolver='hogbom', niter=500, cyclefactor=3, scales=[0, 3, 10] ):
              # use rec to decide when to stop, for example to check stopcode or peak residual:
              # if (rec[0] > 1) or (min(rec[1][0][0]['peakRes']) < 0.001):
              #     break
              print(rec)
    or as an async generator:
          async for rec in gclean( vis='refim_point_withline.ms', imagename='test', imsize=512, cell='12.0arcsec',
                                   specmode='cube', interpolation='nearest', nchan=5, start='1.0GHz', width='0.2GHz',
                                   pblimit=-1e-05, deconvolver='hogbom', niter=500, cyclefactor=3, scales=[0, 3, 10] ):
              # use rec to decide when to stop
              print(rec)


    See also: __next__(...) for a description of the returned rec

    TODO: do we need to preserve any hidden state between tclean calls for the iterbotsink and/or synthesisimager tools?
    '''

    def _tclean( self, *args, **kwargs ):
        """ Calls tclean records the arguments in the local history of tclean calls.

        The full tclean history for this instance can be retrieved via the cmds() method."""
        arg_s = ', '.join( map( lambda a: self._history_filter(len(self._exe_cmds), None, repr(a)), args ) )
        kw_s = ', '.join( map( lambda kv: self._history_filter(len(self._exe_cmds), kv[0], "%s=%s" % (kv[0],repr(kv[1]))), kwargs.items()) )
        if len(arg_s) > 0 and len(kw_s) > 0:
            parameters = arg_s + ", " + kw_s
        else:
            parameters = arg_s + kw_s
        self._exe_cmds.append( "tclean( %s )" % parameters )
        self._exe_cmds_per_iter[-1] += 1
        return tclean( *args, **kwargs )

    def _deconvolve( self, *args, **kwargs ):
        """ Calls deconvolve records the arguments in the local history of deconvolve calls.

        The full deconvolve history for this instance can be retrieved via the cmds() method."""
        arg_s = ', '.join( map( lambda a: self._history_filter(len(self._exe_cmds), None, repr(a)), args ) )
        kw_s = ', '.join( map( lambda kv: self._history_filter(len(self._exe_cmds), kv[0], "%s=%s" % (kv[0],repr(kv[1]))), kwargs.items()) )
        if len(arg_s) > 0 and len(kw_s) > 0:
            parameters = arg_s + ", " + kw_s
        else:
            parameters = arg_s + kw_s
        self._exe_cmds.append( "deconvolve( %s )" % parameters )
        self._exe_cmds_per_iter[-1] += 1
        return deconvolve( *args, **kwargs )

    def _remove_tree( self, directory ):
        if os.path.isdir(directory):
            shutil.rmtree(directory)
            self._exe_cmds.append( f'''shutil.rmtree( {repr(directory)} )''' )
            self._exe_cmds_per_iter[-1] += 1

    def cmds( self, history=False ):
        """ Returns the history of all tclean calls for this instance. If ``history``
        is set to True then the full history will be returned, otherwise the commands
        executed for generating the latest result are returned.
        """

        if history:
            return self._exe_cmds
        else:
            if self._exe_cmds_per_iter[-1] > 0:
                # Return the last N commands
                return self._exe_cmds[-self._exe_cmds_per_iter[-1]:]
            else:
                # If convergence is hit, no commands were run so return nothing
                return []


    def update( self, msg ):
        """ Interactive clean parameters update.

        Args:
            msg: dict with possible keys 'niter', 'cycleniter', 'nmajor', 'threshold', 'cyclefactor'

        Returns:
            stopcode : Stop code in case of error (-1 on error, 0 if no error), int
            stopdesc : Exception error message, str
        """
        if 'niter' in msg:
            try:
                self._niter = int(msg['niter'])
                if self._niter < -1:
                    return -1, f"niter must be >= -1"
            except ValueError as err:
                return -1, "niter must be an integer"

        if 'cycleniter' in msg:
            try:
                self._cycleniter = int(msg['cycleniter'])
                if self._cycleniter < -1:
                    return -1, f"cycleniter must be >= -1"
            except ValueError:
                return -1, "cycleniter must be an integer"

        if 'nmajor' in msg:
            try:
                self._nmajor = int(msg['nmajor'])
                if self._nmajor < -1:
                    return -1, f"nmajor must be >= -1"
            except ValueError as e:
                return -1, "nmajor must be an integer"

        if 'threshold' in msg:
            try:
                self._threshold = float(msg['threshold'])
                if self._threshold < 0:
                    return -1, f"threshold must be >= 0"
            except ValueError:
                if isinstance(msg['threshold'], str) and "jy" in msg['threshold'].lower():
                    self._threshold_to_float(msg['threshold']) # Convert str to float
                else:
                    return -1, f"threshold must be a number, or a number with units (Jy/mJy/uJy)"


        if 'nsigma' in msg:
            try:
                self._nsigma = float(msg['nsigma'])
                if self._nsigma < 0:
                    return -1, f"nsigma must be >= 0"
            except ValueError:
                return -1, "nsigma must be a number"


        if 'gain' in msg:
            try:
                self._gain = float(msg['gain'])
                if self._gain <= 0:
                    return -1, f"gain must be > 0"
            except ValueError:
                return -1, "gain must be a number"


        if 'cyclefactor' in msg:
            try:
                self._cyclefactor = float(msg['cyclefactor'])
                if self._cyclefactor <= 0:
                    return -1, f"cyclefactor must be > 0"
            except ValueError:
                return -1, "cyclefactor must be a number"

        ### Automasking parameters

        if 'dogrowprune' in msg:
            try:
                self._dogrowprune = bool(msg['dogrowprune'])
            except ValueError:
                return -1, "dogrowprune must be a boolean"

        if 'noisethreshold' in msg:
            try:
                self._noisethreshold = float(msg['noisethreshold'])
                if self._noisethreshold < 0:
                    return -1, f"noisethreshold must be >= 0"
            except ValueError:
                return -1, "noisethreshold must be a number"

        if 'sidelobethreshold' in msg:
            try:
                self._sidelobethreshold = float(msg['sidelobethreshold'])
                if self._sidelobethreshold < 0:
                    return -1, f"sidelobethreshold must be >= 0"
            except ValueError:
                return -1, "sidelobethreshold must be a number"

        if 'lownoisethreshold' in msg:
            try:
                self._lownoisethreshold = float(msg['lownoisethreshold'])
                if self._lownoisethreshold < 0:
                    return -1, f"lownoisethreshold must be >= 0"
            except ValueError:
                return -1, "lownoisethreshold must be a number"

        if 'minbeamfrac' in msg:
            try:
                self._minbeamfrac = float(msg['minbeamfrac'])
                if self._minbeamfrac < 0:
                    return -1, f"minbeamfrac must be >= 0"
            except ValueError:
                return -1, "minbeamfrac must be a number"

        if 'negativethreshold' in msg:
            try:
                self._negativethreshold = float(msg['negativethreshold'])
                if self._negativethreshold < 0:
                    return -1, f"negativethreshold must be >= 0"
            except ValueError:
                return -1, "negativethreshold must be a number"



        return 0, ""



    def _threshold_to_float(self, msg=None):
        # Convert threshold from string to float if necessary
        if msg is not None:
            if isinstance(msg, str):
                if "mJy" in msg:
                    self._threshold = float(msg.replace("mJy", "")) / 1e3
                elif "uJy" in msg:
                    self._threshold = float(msg.replace("uJy", "")) / 1e6
                elif "Jy" in msg:
                    self._threshold = float(msg.replace("Jy", ""))
        else:
            if isinstance(self._threshold, str):
                if "mJy" in self._threshold:
                    self._threshold = float(self._threshold.replace("mJy", "")) / 1e3
                elif "uJy" in self._threshold:
                    self._threshold = float(self._threshold.replace("uJy", "")) / 1e6
                elif "Jy" in self._threshold:
                    self._threshold = float(self._threshold.replace("Jy", ""))


    def __init__(self, vis, imagename, selectdata=True, field='', spw='', timerange='', uvrange='', antenna='', scan='', observation='', intent='', datacolumn='corrected', imsize=[100], cell=[ ],
                 phasecenter='', projection='SIN', stokes='I', startmodel='', specmode='cube', reffreq='', nchan=-1, start='', width='', outframe='LSRK', veltype='radio', restfreq='',
                 interpolation='linear', perchanweightdensity=True, gridder='standard', facets=int(1), psfphasecenter='', wprojplanes=int(1), mosweight=True, aterm=True, psterm=False, wbawp=True,
                 conjbeams=False, usepointing=False, cfcache = '', pointingoffsetsigdev=[  ], vptable='', computepastep=float(360.0), rotatepastep=float(360.0), pblimit=0.2, normtype='flatnoise',
                 deconvolver='hogbom', smallscalebias=0.0, fusedthreshold=0, largestscale=-1, niter=0, threshold='0.1Jy', nsigma=0.0, cycleniter=-1, nmajor=1, cyclefactor=1.0, minpsffraction=0.05,
                 maxpsffraction=0.8, scales=[], restoringbeam='', pbcor=False, outlierfile='', nterms=int(2), weighting='natural', robust=float(0.5), noise='0.0Jy', uvtaper=[], npixels=0,
                 gain=float(0.1), pbmask=0.2, sidelobethreshold=3.0, noisethreshold=5.0, lownoisethreshold=1.5, negativethreshold=0.0, smoothfactor=float(1.0), minbeamfrac=0.3, cutthreshold=0.01,
                 growiterations=75, dogrowprune=True, minpercentchange=-1.0, verbose=False, fastnoise=True, savemodel='none', usemask='user', mask='', restoration=True, restart=True, calcres=True,
                 calcpsf=True, psfcutoff=float(0.35), parallel=False, history_filter=lambda index, arg, history_value: history_value ):

        self._vis = vis
        self._imagename = imagename
        self._selectdata = selectdata
        self._imsize = imsize
        self._cell = cell
        self._phasecenter = phasecenter
        self._projection = projection
        self._stokes = stokes
        self._startmodel = startmodel
        self._specmode = specmode
        self._reffreq = reffreq
        self._nchan = nchan
        self._start = start
        self._width = width
        self._outframe = outframe
        self._veltype = veltype
        self._restfreq = restfreq
        self._interpolation = interpolation
        self._perchanweightdensity = perchanweightdensity
        self._gridder = gridder
        self._facets = facets
        self._psfphasecenter = psfphasecenter
        self._wprojplanes = wprojplanes
        self._mosweight = mosweight
        self._aterm = aterm
        self._psterm = psterm
        self._wbawp = wbawp
        self._conjbeams = conjbeams
        self._usepointing = usepointing
        self._cfcache = cfcache
        self._pointingoffsetsigdev = pointingoffsetsigdev
        self._vpable = vptable
        self._computepastep = computepastep
        self._rotatepastep = rotatepastep
        self._pblimit = pblimit
        self._normtype = normtype
        self._deconvolver = deconvolver
        self._smallscalebias = smallscalebias
        self._fusedthreshold = fusedthreshold
        self._largestscale = largestscale
        self._niter = niter
        self._threshold = threshold
        self._cycleniter = cycleniter
        self._minpsffraction = minpsffraction
        self._maxpsffraction = maxpsffraction
        self._nsigma = nsigma
        self._nmajor = nmajor
        self._cyclefactor = cyclefactor
        self._scales = scales
        self._restoringbeam = restoringbeam
        self._pbcor = pbcor
        #self._outlierfile = outlierfile
        self._nterms = nterms
        self._exe_cmds = [ ]
        self._exe_cmds_per_iter = [ ]
        self._history_filter = history_filter
        self._finalized = False
        self._field = field
        self._spw = spw
        self._timerange = timerange
        self._uvrange = uvrange
        self._antenna = antenna
        self._scan = scan
        self._observation = observation
        self._intent = intent
        self._datacolumn = datacolumn
        self._weighting = weighting
        self._robust = robust
        self._noise = noise
        self._uvtaper = uvtaper
        self._npixels = npixels
        self._gain = gain
        self._pbmask = pbmask
        self._sidelobethreshold = sidelobethreshold
        self._noisethreshold = noisethreshold
        self._lownoisethreshold = lownoisethreshold
        self._negativethreshold = negativethreshold
        self._smoothfactor = smoothfactor
        self._minbeamfrac = minbeamfrac
        self._cutthreshold = cutthreshold
        self._growiterations = growiterations
        self._dogrowprune = dogrowprune
        self._minpercentchange = minpercentchange
        self._verbose = verbose
        self._fastnoise = fastnoise
        self._savemodel = savemodel
        self._parallel = parallel
        self._usemask = usemask
        self._mask = mask
        self._restoration = restoration
        self._restart = restart
        self._calcres = calcres
        self._calcpsf = calcpsf
        self._psfcutoff = psfcutoff
        self.global_imdict = ImagingDict()
        self.current_imdict = ImagingDict()
        self._major_done = 0
        self.hasit = False # Convergence flag
        self._has_restored = False
        self.stopdescription = '' # Convergence flag
        self._initial_mask_exists = False
        self._convergence_result = (None,None,None,None,None,{ 'chan': None, 'major': None })
        #                           ^^^^ ^^^^ ^^^^ ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^----->>> convergence info
        #                              |    | |     |    +----->>> Number of global iterations remaining for current run (niterleft)
        #                              |    | |     +---------->>> Number of major cycles remaining for current run (nmajorleft)
        #                              |    | +---------------->>> major cycles done for current run (nmajordone)
        #                              |    +------------------>>> tclean stopcode
        #                              +----------------------->>> tclean stopdescription

        # Convert threshold from string to float, interpreting units.
        # XXX : We should ideally use quantities, but we are trying to
        # stick to "public API" funtions inside _gclean
        self._threshold_to_float()


    def __add_per_major_items( self, tclean_ret, major_ret, chan_ret ):
        '''Add meta-data about the whole major cycle, including 'cyclethreshold'
        '''

        if 'cyclethreshold' in tclean_ret:

            rdict = dict( major=dict( cyclethreshold=[tclean_ret['cyclethreshold']] if major_ret is None else (major_ret['cyclethreshold'] + [tclean_ret['cyclethreshold']]) ),
                         chan=chan_ret )
        else:
            rdict = dict( major=dict( cyclethreshold=major_ret['cyclethreshold'].append(tclean_ret['cyclethreshold']) ),
                         chan=chan_ret )

        return rdict


    def _calc_deconv_controls(self, imdict, niterleft=0, threshold=0, cycleniter=-1):
        """
        Calculate cycleniter and cyclethreshold for deconvolution.
        """

        use_cycleniter = niterleft  #niter - imdict.returndict['iterdone']

        if cycleniter > -1 : # User is forcing this number
            use_cycleniter = min(cycleniter, use_cycleniter)

        psffrac = imdict.returndict['maxpsfsidelobe'] * self._cyclefactor
        psffrac = max(psffrac, self._minpsffraction)
        psffrac = min(psffrac, self._maxpsffraction)

        # TODO : This assumes the default field (i.e., field=0);
        # This won't work for multiple fields.
        cyclethreshold = psffrac * imdict.get_peakres()
        cyclethreshold = max(cyclethreshold, threshold)

        return int(use_cycleniter), cyclethreshold


    def __update_convergence(self):
        """
        Accumulates the per-channel/stokes summaryminor keys across all major cycle calls so far.

        The "iterDone" key will be replaced with "iterations", and for the "iterations" key,
        the value in the returned cummulative record will be a rolling sum of iterations done
        for tclean calls so far, one value per minor cycle.
        For example, if there have been two clean calls, and in the first call channel 0 had
        [1] iteration in 1 minor cycle, and for the second call channel 0 had [6, 10, 9, 1]
        iterations in 4 minor cycles), then the resultant "iterations" key for channel 0 would be:
        [1, 7, 17, 26, 27]
        """

        keys = ['modelFlux', 'iterDone', 'peakRes', 'stopCode', 'cycleThresh']

        # Grab tuples of keys of interest
        outrec = {}
        for nn in range(self.global_imdict.nchan):
            outrec[nn] = {}
            for ss in range(self.global_imdict.nstokes):
                outrec[nn][ss] = {}
                for key in keys:
                    # Replace iterDone with iterations
                    if key == 'iterDone':
                        # Maintain cumulative sum of iterations per entry
                        outrec[nn][ss]['iterations'] = np.cumsum(self.global_imdict.get_key(key, stokes=ss, chan=nn))
                        # Replace iterDone with iterations
                        #outrec[nn][ss]['iterations'] =  self.global_imdict.get_key(key, stokes=ss, chan=nn)
                    else:
                        outrec[nn][ss][key] = self.global_imdict.get_key(key, stokes=ss, chan=nn)

        return outrec


    def _check_initial_mask(self):
        """
        Check if a mask from a previous run exists on disk or not.
        """

        if self._usemask == 'user' and self._mask == '':
            maskname = self._imagename + '.mask'

            if os.path.exists(maskname):
                self._initial_mask_exists = True
            else:
                self._initial_mask_exists = False

    def _fix_initial_mask(self):
        """
        If on start up, no user mask is provided, then flip the initial mask to
        be all zeros for interactive use.
        """

        from casatools import image
        ia = image()

        if self._usemask == 'user' and self._mask == '':
            maskname = self._imagename + '.mask'

            # This means the mask was newly created by deconvolve, so flip it
            if os.path.exists(maskname) and self._initial_mask_exists is False:
                ia.open(maskname)
                ia.set(0.0)
                ia.close()

    def _update_peakres(self):
        if self._deconvolver == 'mtmfs':
            residname = self._imagename + '.residual.tt0'
        else:
            residname = self._imagename + '.residual'

        maskname = self._imagename + '.mask'
        if not os.path.exists(maskname):
            maskname = ''

        peakres = imstat(imagename=residname, mask=f'''"{maskname}"''')['max']
        if len(maskname) > 0:
            masksum = imstat(imagename=maskname)['sum']
        else:
            masksum = []

        if len(peakres) > 0:
            peakres = peakres[0]
        else:
            peakres = None

        if len(masksum) > 0:
            masksum = masksum[0]
        else:
            masksum = None

        return peakres, masksum

    def __next__( self ):
        """ Runs tclean and returns the (stopcode, convergence result) when executed with the python builtin next() function.

        The returned convergence result is a nested dictionary:
        {
            channel id: {
                stokes id: {
                    summary key: [values, one per minor cycle]
                },
            },
        }

        See also: gclean.__update_convergence(...)
        """

        tclean_ret = {}
        deconv_ret = {}

        #                      vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv------------>>> is done to produce an initial dirty image
        if self._niter < 1 and self._convergence_result[2] is not None:
            self._convergence_result = ( f'nothing to run, niter == {self._niter}',
                                         self._convergence_result[1],
                                         self._major_done,
                                         self._nmajor,
                                         self._niter,
                                         self._convergence_result[5] )
            return self._convergence_result
        else:
            ### CALL SEQUENCE:
            ###       tclean(niter=0),deconvolve(niter=0),tclean(niter=100),deconvolve(niter=0),tclean(niter=100),tclean(niter=0,restoration=True)
            self._exe_cmds_per_iter.append(0)
            try:
                if self._convergence_result[1] is None:
                    # initial call to tclean(...) creates the initial dirty image with niter=0

                    # If calcres and calcpsf are False, no need to run initial tclean - assume image products already on disk.
                    if not (self._calcres == False and self._calcpsf == False):
                        casalog.post('Running initial major cycle to create first residual image.', 'INFO')
                        print('Running initial major cycle to create first residual image.')
                        tclean_ret = self._tclean( vis=self._vis, mask=self._mask, imagename=self._imagename, imsize=self._imsize, cell=self._cell, selectdata=self._selectdata, phasecenter=self._phasecenter, stokes=self._stokes,
                                              startmodel=self._startmodel, specmode=self._specmode, projection=self._projection, reffreq=self._reffreq, gridder=self._gridder, wprojplanes=self._wprojplanes, facets=self._facets,
                                              mosweight=self._mosweight, psfphasecenter = self._psfphasecenter, aterm=self._aterm, psterm=self._psterm, wbawp=self._wbawp, conjbeams=self._conjbeams, cfcache=self._cfcache,
                                              usepointing=self._usepointing, computepastep=self._computepastep, rotatepastep=self._rotatepastep, normtype=self._normtype, fusedthreshold=self._fusedthreshold,
                                              largestscale=self._largestscale, noise = self._noise, uvtaper=self._uvtaper, psfcutoff=self._psfcutoff, interpolation=self._interpolation,
                                              perchanweightdensity=self._perchanweightdensity, nchan=self._nchan, start=self._start, width=self._width, veltype=self._veltype, restfreq=self._restfreq, outframe=self._outframe,
                                              pointingoffsetsigdev=self._pointingoffsetsigdev, pblimit=self._pblimit, deconvolver=self._deconvolver, smallscalebias=self._smallscalebias, cyclefactor=self._cyclefactor,
                                              scales=self._scales, restoringbeam=self._restoringbeam, pbcor=self._pbcor, nterms=self._nterms, field=self._field, spw=self._spw, timerange=self._timerange, uvrange=self._uvrange,
                                              antenna=self._antenna, scan=self._scan, observation=self._observation, intent=self._intent, datacolumn=self._datacolumn, weighting=self._weighting, robust=self._robust,
                                              npixels=self._npixels, interactive=False, niter=0, gain=self._gain, calcres=self._calcres, calcpsf=self._calcpsf, restoration=False, parallel=self._parallel, fullsummary=True)
                                              # outlierfile = self._outlierfile,

                    # Check if a mask from a previous run exists on disk
                    self._check_initial_mask()

                    deconv_ret = self._deconvolve(imagename=self._imagename, startmodel=self._startmodel,
                                                  deconvolver=self._deconvolver, scales=self._scales, nterms=self._nterms,
                                                  smallscalebias=self._smallscalebias, restoration=False, restoringbeam=self._restoringbeam,
                                                  niter = 0, gain=self._gain, threshold=self._threshold, nsigma=self._nsigma,
                                                  interactive = False, fullsummary=True, fastnoise=self._fastnoise, usemask=self._usemask,
                                                  mask = self._mask, pbmask=self._pbmask, sidelobethreshold=self._sidelobethreshold,
                                                  noisethreshold=self._noisethreshold, lownoisethreshold=self._lownoisethreshold,
                                                  negativethreshold=self._negativethreshold, smoothfactor=self._smoothfactor,
                                                  minbeamfrac=self._minbeamfrac, cutthreshold=self._cutthreshold,
                                                  growiterations=self._growiterations, dogrowprune=self._dogrowprune,
                                                  verbose=self._verbose)

                    # If no mask from a previous run exists, over-write the ones with zeros for the default mask
                    self._fix_initial_mask()


                    if len(tclean_ret) > 0 and len(deconv_ret) > 0:
                        self.current_imdict.returndict = self.current_imdict.merge(tclean_ret, deconv_ret)
                    elif len(tclean_ret) > 0 and len(deconv_ret) == 0:
                        self.current_imdict.returndict = copy.deepcopy(tclean_ret)
                    elif len(tclean_ret) == 0 and len(deconv_ret) > 0:
                        self.current_imdict.returndict = copy.deepcopy(deconv_ret)
                    else: # Both dicts are empty, this should never happen
                        raise ValueError("Both tclean and deconvolve return dicts are empty. This should never happen.")

                    self.global_imdict.returndict = self.current_imdict.returndict

                    ## Initial call where niterleft and nmajorleft are same as original input values.
                    self.hasit, self.stopdescription = self.global_imdict.has_converged(self._niter, self._threshold, self._nmajor)

                    self.current_imdict.returndict['stopcode'] = self.hasit
                    self.current_imdict.returndict['stopDescription'] = self.stopdescription
                    self._major_done = 0
                else:
                    # Reset convergence every time, since we return control to the GUI after a single major cycle
                    self.current_imdict.returndict['iterdone'] = 0.

                    # Mask can be updated here...
                    # Check for mask update - peakres + masksum
                    _peakres, _masksum = self._update_peakres()

                    self.hasit, self.stopdescription = self.global_imdict.has_converged(self._niter, self._threshold, self._nmajor, masksum=_masksum, peakres=_peakres)

                    #self.global_imdict.returndict['stopcode'] = self.hasit
                    #self.global_imdict.returndict['stopDescription'] = self.stopdescription

                    #self.current_imdict.returndict['stopcode'] = self.hasit
                    #self.current_imdict.returndict['stopDescription'] = self.stopdescription

                    # Has not, i.e., not converged
                    if self.hasit ==0 :
                        use_cycleniter, cyclethreshold = self._calc_deconv_controls(self.current_imdict, self._niter, self._threshold, self._cycleniter)

                        # Run the minor cycle
                        deconv_ret = self._deconvolve(imagename=self._imagename, startmodel=self._startmodel,
                                                  deconvolver=self._deconvolver, restoration=False,
                                                  threshold=cyclethreshold, niter=use_cycleniter, gain=self._gain, fullsummary=True)

                        # Run the major cycle
                        tclean_ret = self._tclean( vis=self._vis, imagename=self._imagename, imsize=self._imsize, cell=self._cell,
                                           phasecenter=self._phasecenter, stokes=self._stokes, specmode=self._specmode, reffreq=self._reffreq,
                                           gridder=self._gridder, wprojplanes=self._wprojplanes, mosweight=self._mosweight, psterm=self._psterm,
                                           wbawp=self._wbawp, conjbeams=self._conjbeams, usepointing=self._usepointing, interpolation=self._interpolation,
                                           perchanweightdensity=self._perchanweightdensity, nchan=self._nchan, start=self._start,
                                           width=self._width, veltype=self._veltype, restfreq=self._restfreq, outframe=self._outframe,
                                           pointingoffsetsigdev=self._pointingoffsetsigdev, pblimit=self._pblimit, deconvolver=self._deconvolver,
                                           smallscalebias=self._smallscalebias, cyclefactor=self._cyclefactor, scales=self._scales,
                                           restoringbeam=self._restoringbeam, pbcor=self._pbcor, nterms=self._nterms, field=self._field,
                                           spw=self._spw, timerange=self._timerange, uvrange=self._uvrange, antenna=self._antenna,
                                           scan=self._scan, observation=self._observation, intent=self._intent, datacolumn=self._datacolumn,
                                           weighting=self._weighting, robust=self._robust, npixels=self._npixels, interactive=False,
                                           niter=0, restart=True, calcpsf=False, calcres=True, restoration=False, threshold=self._threshold,
                                           nsigma=self._nsigma, cycleniter=self._cycleniter, nmajor=1, gain=self._gain,
                                           sidelobethreshold=self._sidelobethreshold, noisethreshold=self._noisethreshold,
                                           lownoisethreshold=self._lownoisethreshold, negativethreshold=self._negativethreshold,
                                           minbeamfrac=self._minbeamfrac, growiterations=self._growiterations, dogrowprune=self._dogrowprune,
                                           minpercentchange=self._minpercentchange, fastnoise=self._fastnoise, savemodel=self._savemodel,
                                           maxpsffraction=self._maxpsffraction,
                                           minpsffraction=self._minpsffraction, parallel=self._parallel, fullsummary=True )

                        # Replace return dict with new return dict
                        # The order of the dicts into merge is important.
                        self.current_imdict.returndict = self.current_imdict.merge(tclean_ret, deconv_ret)

                        # Append new return dict to global return dict
                        self.global_imdict.returndict = self.global_imdict.concat(self.global_imdict.returndict, self.current_imdict.returndict)
                        self._major_done = self.current_imdict.returndict['nmajordone']

                        ## Decrement count for the major cycle just done...
                        self.__decrement_counts()

                        cycleniterleft = self._cycleniter - self.current_imdict.returndict['iterdone']

                        # Use global imdict for convergence check
                        if deconv_ret['stopcode'] == 7:   ## Tell the convergence checker that the mask is zero and iterations were skipped
                            self.hasit, self.stopdescription = self.global_imdict.has_converged(self._niter, self._threshold, self._nmajor, masksum=0)
                        else:
                            self.hasit, self.stopdescription = self.global_imdict.has_converged(self._niter, self._threshold, self._nmajor, cycleniter=cycleniterleft)


                    self.global_imdict.returndict['stopcode'] = self.hasit
                    self.global_imdict.returndict['stopDescription'] = self.stopdescription

                    if not self.hasit and self._usemask == 'auto-multithresh':
                        # If we haven't converged, run deconvolve to update the mask
                        # Note : This is only necessary if using auto-multithresh. If using the interactive viewer to draw, the mask gets updated as the regions
                        # are added or removed in the interactive viewer.
                        self._deconvolve(imagename=self._imagename, startmodel=self._startmodel, deconvolver=self._deconvolver, restoration=False, threshold=self._threshold, niter=0,
                                         nsigma=self._nsigma, fullsummary=True, fastnoise=self._fastnoise, usemask=self._usemask, mask='', pbmask=self._pbmask,
                                         sidelobethreshold=self._sidelobethreshold, noisethreshold=self._noisethreshold, lownoisethreshold=self._lownoisethreshold,
                                         negativethreshold=self._negativethreshold, smoothfactor=self._smoothfactor, minbeamfrac=self._minbeamfrac, cutthreshold=self._cutthreshold,
                                         growiterations=self._growiterations, dogrowprune=self._dogrowprune, verbose=self._verbose)


                if len(self.global_imdict.returndict) > 0 and 'summaryminor' in self.global_imdict.returndict and sum(map(len,self.global_imdict.returndict['summaryminor'].values())) > 0:
                    self._convergence_result = ( self.global_imdict.returndict['stopDescription'] if 'stopDescription' in self.global_imdict.returndict else '',
                                                 self.global_imdict.returndict['stopcode'] if 'stopcode' in self.global_imdict.returndict else 0,
                                                 self._major_done,
                                                 self._nmajor,
                                                 self._niter,
                                                 self.__add_per_major_items( self.global_imdict.returndict,
                                                                             self._convergence_result[5]['major'],
                                                                             self.__update_convergence()))
                else:
                    self._convergence_result = ( f'tclean returned an empty result',
                                                 self._convergence_result[1],
                                                 self._major_done,
                                                 self._nmajor,
                                                 self._niter,
                                                 self._convergence_result[5] )
            except Exception as e:
                self._convergence_result = ( str(e),
                                             -1,
                                             self._major_done,
                                             self._nmajor,
                                             self._niter,
                                             self._convergence_result[5] )
                return self._convergence_result

            return self._convergence_result

    def __decrement_counts( self ):
        ## Update niterleft and nmajorleft now.
        if self.hasit == 0:  ##If not yet converged.
            if self._nmajor != -1:   ## If -1, don't touch it.
                self._nmajor = self._nmajor - 1
                if self._nmajor<0:   ## Force a floor
                    self._nmajor=0
            self._niter = self._niter - self.current_imdict.get_key('iterdone')
            if self._niter<0:  ## This can happen when we're counting niter across channels in a single minor cycle set, and it crosses the total.
                self._niter=0  ## Force a floor
        else:
            return  ##If convergence has been reached, don't try to decrement further.


    def __reflect_stop( self ):
        ## if python wasn't hacky, you would be able to try/except/raise in lambda
        time.sleep(1)
        try:
            return self.__next__( )
        except StopIteration:
            raise StopAsyncIteration

    async def __anext__( self ):
        ### asyncio.run cannot be used here because this is called
        ### within an asyncio loop...
        loop = asyncio.get_event_loop( )
        result = await loop.run_in_executor( None, self.__reflect_stop )
        return result

    def __iter__( self ):
        return self

    def __aiter__( self ):
        return self

    def __split_filename( self, path ):
        return os.path.splitext(os.path.basename(path))

    def __default_mask_name( self ):
        imgparts = self.__split_filename( self._imagename )
        return f'{imgparts[0]}.mask'

    def __del__(self):
        if not self._has_restored:
            self.restore()

    def mask(self):
        #return self.__default_mask_name() if self._mask == '' else self._mask
        return f'{self._imagename}.mask' if self._mask == '' else self._mask

    def reset(self):
        #if not self._finalized:
        #    raise RuntimeError('attempt to reset a gclean run that has not been finalized')
        self._finalized = False
        self._convergence_result = ( None,
                                     self._convergence_result[1],
                                     self._major_done,
                                     self._nmajor,
                                     self._niter,
                                     self._convergence_result[5] )


    def restore(self):
        """ Restores the final image, and returns a path to the restored image. """
        deconv_ret = self._deconvolve(imagename=self._imagename,
                                      deconvolver=self._deconvolver,
                                      restoration=True, niter=0,
                                      fullsummary=True)

        self._has_restored = True

        return { "image": f"{self._imagename}.image" }

    def has_next(self):
        return not self._finalized



from bokeh.models import PolyAnnotation
from .. import BokehInit

class EvPolyAnnotation(PolyAnnotation,BokehInit):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)



########################################################################
#
# Copyright (C) 2025
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
import hashlib
from astropy.wcs import WCS
from regions import PixCoord, PolygonPixelRegion

class _Line:

    def __init__( self, style ):
        if not ( 'color' in style and 'width' in style and 'alpha' in style and 'dash' in style ):
            raise RuntimeError( 'style must contain color, width, alpha and dash' )
        self.__style = style

    def color( self ):
        return self.__style['color']

    def width( self ):
        return self.__style['width']

    def alpha( self ):
        return self.__style['alpha']

    def dash( self ):
        return self.__style['dash']

class _Fill:

    def __init__( self, fill ):
        if not ( 'color' in fill and 'alpha' in fill and 'hatch' in fill ):
            raise  RuntimeError( 'fill must contain color, alpha and hatch' )
        self.__fill = fill

    def color( self ):
        return self.__fill['color']

    def alpha( self ):
        return self.__fill['alpha']

    def hatch( self ):
        return self.__fill['hatch']

class Region:

    ### reuse astropy WCS objects based upon fits header strings
    ### to avoid reparsing
    wcs_cache = { }

    def __init__( self, name, xs, ys, channels, line, fill, wcs=None ):
        if type(name) != str:
            raise RuntimeError( 'region names are strings' )
        if len(xs) != len(ys):
            raise RuntimeError( 'X and Y values must be of equal length' )

        self.__ste = { }
        self.__ste['name'] = name
        self.__ste['xs'] = xs
        self.__ste['ys'] = ys
        self.__ste['chan'] = channels
        self.__ste['line'] = line
        self.__ste['fill'] = fill
        self.__ste['wcs'] = None
        if wcs is not None:
            if isinstance(wcs,WCS):
                self.__ste['wcs'] = wcs
            elif type(wcs) == str:
                ## Assume wcs is a FITS header
                hash = hashlib.md5(wcs.encode()).hexdigest()
                if hash not in Region.wcs_cache:
                    Region.wcs_cache[hash] = WCS( header=wcs )
                self.__ste['wcs'] = Region.wcs_cache[hash]
            else:
                raise RuntimeError( f'''Regions initialized with non-FITS header representations of the WCS of type {type(wcs)} are not supported''' )

    def __str__(self):
        return self.__repr__( )

    def __repr__(self):
         return f'''<{self.__class__.__name__}({self.__ste['name']}) @ {hex(id(self))}>'''

    def name( self ):
        return self.__ste['name']

    def xs( self ):
        return self.__ste['xs']

    def ys( self ):
        return self.__ste['ys']

    def vertices( self ):
        return zip( self.__ste['xs'], self.__ste['ys'] )

    def channels( self ):
        return self.__ste['chan']

    def line( self ):
        return _Line( self.__ste['line'] )

    def fill( self ):
        return _Fill( self.__ste['fill'] )

    def to_astropy_pixel( self ):
        ### It seems as though there is no way to create 4 dimensional pixel coordinates (RA,DEC,STOKES,CHAN)
        ### using astropy.
        return PolygonPixelRegion( vertices=PixCoord( self.__ste['xs'], self.__ste['ys'],
                                                      ### meta and visual are not available on regions v0.10 which
                                                      ### seems to be the most recent version available with pip
                                                      #meta={ 'name': self.__ste['name'] },
                                                      #visual={ 'dash': self.__ste['line']['dash'],
                                                      #         'linewidth': self.__ste['line']['width'] }
                                                     ) )
                                                             

    def to_astropy_sky( self, wcs=None ):
        if wcs is None:
            wcs = self.__ste['wcs']
        if wcs is None:
            raise RuntimeError( 'No world coordinate system is available' )
        ### Astropy coordinates seem to be limited to two dimensions so the generalized
        ### WCS object (potentially with stokes and frequency axes) must be converted to
        ### a two coordinate description.
        return self.to_astropy_pixel( ).to_sky( wcs.sub([1,2]) )

class _Iter:
    def __init__( self, ste ):
        self.__ste = ste

    def __iter__(self):
        for name, rgn in self.__ste['raw'].items( ):
            yield ( name, Region( name, rgn['geometry']['xs'], rgn['geometry']['ys'],
                                    rgn['channels'], rgn['styling']['line'],
                                    rgn['styling']['fill'], wcs=self.__ste['wcs'] ) )

class RegionList:
    '''This class is for representing regions within cubevis applications. Currently it only wowrks with the
    `createregion` app. It is not (currently) designed as an end user class but rather as an output class for
    the user.

    Currently it only accepts a dictionary like for `regiondict`:
    ```
    {'rg0': {'channels': [[0, 0]], 'geometry': {'xs': [33.51816859449925, 31.611101128190022, 121.40193309213468], 'ys': [195.13666993774888, 232.39613665715922, 229.12483086356568]}, 'styling': {'line': {'color': '#ffffff', 'width': 1, 'alpha': 1, 'dash': 'solid'}, 'fill': {'color': '#ffffff', 'alpha': 0, 'hatch': 'blank'}}}, 'rg1': {'channels': [[0, 0]], 'geometry': {'xs': [289.43821781788444, 347.2289277115166, 351.7707259199244], 'ys': [236.68983013019135, 232.27134277978703, 186.01834737432947]}, 'styling': {'line': {'color': '#ffffff', 'width': 1, 'alpha': 1, 'dash': 'solid'}, 'fill': {'color': '#ffffff', 'alpha': 0, 'hatch': 'blank'}}}, 'rg2': {'channels': [[0, 0]], 'geometry': {'xs': [346.576046788646, 334.82618711394673, 270.71434943666696], 'ys': [61.16382772497764, 9.769953019597281, 16.587407740876944]}, 'styling': {'line': {'color': '#ffffff', 'width': 1, 'alpha': 1, 'dash': 'solid'}, 'fill': {'color': '#ffffff', 'alpha': 0, 'hatch': 'blank'}}}, 'rg3': {'channels': [[0, 0]], 'geometry': {'xs': [17.33383873194185, 20.60032838342714, 105.68674392600984], 'ys': [65.03236450671528, 26.158256778791444, 15.71542171880079]}, 'styling': {'line': {'color': '#ffffff', 'width': 1, 'alpha': 1, 'dash': 'solid'}, 'fill': {'color': '#ffffff', 'alpha': 0, 'hatch': 'blank'}}}, 'rg4': {'channels': [[0, 0]], 'geometry': {'xs': [194.1150775408404, 186.7420925740613, 247.32924839610087, 257.7778686999655, 220.3835500015239], 'ys': [100.28458653562863, 138.0450054510982, 136.91400495710235, 109.48993737427867, 75.12723329938433]}, 'styling': {'line': {'color': '#ffffff', 'width': 1, 'alpha': 1, 'dash': 'solid'}, 'fill': {'color': '#ffffff', 'alpha': 0, 'hatch': 'blank'}}}}
    ```
    The `wcs` parameter could either be a string representation of a FITS wcs header or it could be an
    astropy `WCS` object.
    '''

    def __init__( self, regiondict, wcs ):

        self.__ste = { 'wcs': wcs }
        self.__summary_size = 22

        if type(regiondict) != dict:
            raise RuntimeError( 'RegionList must be initialized with a dictionary representation of regions' )
        self.__ste['raw'] = regiondict

    def __iter__(self):
        return _Iter(self.__ste).__iter__( )

    def __str__(self):
        return self.__repr__( )

    def __repr__(self):
        added = 0
        names = ''
        keys = self.__ste['raw'].keys( )
        for name in keys:
            if len(names) + len(names) > self.__summary_size: break
            names = names + (',' if len(names) > 0 else '') + name
            added = added + 1
        if added < len(keys):
            names = f'''{names}...'''
        return f'''<{self.__class__.__name__}[{names}] @ {hex(id(self))}>'''
    



########################################################################
#
# Copyright (C) 2021, 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Implementation of a general purpose event manager between Python and
JavaScript via ``websockets``. This provides a mechanism like the
``ImagePipe``. The difference is that ``ImagePipe`` is tuned for use
with CASA images but ``DataPipe`` can have generic messages.'''

import inspect
import threading
import asyncio
import traceback
import time
import json
from uuid import uuid4

from bokeh.models.sources import DataSource
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Tuple, String, Int, Instance, Nullable, Bool
from bokeh.models.callbacks import Callback

from ...utils import serialize, deserialize, is_interactive_jupyter, is_colab
from ..state import casalib_url, cubevisjs_url
from .. import BokehInit

class DataPipe(DataSource,BokehInit):
    """This class allows for communication between Python and the JavaScript implementation
    running in a browser. It allows Python code to send a message to JavaScript and register
    a callback which will receive the result. JavaScript code can do the same to request
    information from Python. Generally messages are expected to be dictionaries in the
    Python domain and objects in the JavaScript domain. UUIDs are used to keep messages in
    sync, and messages sent with the same UUID are queued until the currently pending
    message reply is recieved. A class can use multipe UUIDs to control queuing behavior.

    Attributes
    ----------
    address: tuple of string and int
        the string is the IP address for the network that should be used and the
        integer is the port number, see ``cubevis.utils.find_ws_address``
    init_script: JavaScript
        this javascript is run when this DataPipe object is initialized. init_script
        is used to run caller JavaScript which needs to be run at initialization time.
        This is optional and does not need to be set.
    """

    init_script = Nullable(Instance(Callback), help="""
    JavaScript to be run during initialization of an instance of an DataPipe object.
    """)

    backend_ip = String(default="127.0.0.1", help="the IP address that the Python server binds to locally" )
    backend_port = Int( help="the port that the Python server listens on" )
    backend_url = String( help="the full URL used by the browser to connect to the Python server" )

    instance_id = String( help="Unique ID for each DataPipe object" )

    conflict_check = Bool( default=True, help="Perform check to avoid reuse of URL for GUI. Not needed in the Jupyter context" )

    # Class-level session tracking to prevent multiple connections
    _active_sessions = {}  # session_id -> {'websocket': ws, 'timestamp': time, 'datapipe': instance}
    _session_lock = threading.Lock()

    ###################################################################
    ### filled from cubevis.bokeh.state._initialize._order_bokeh_js ###
    ###################################################################
    #__javascript__ = [ casalib_url( ), cubevisjs_url( ) ]

    def __init__( self, *args, abort=None, address=None, **kwargs ):

        if 'conflict_check' not in kwargs:
            kwargs['conflict_check'] = not is_interactive_jupyter( )
        if 'instance_id' not in kwargs:
            kwargs['instance_id'] = str(uuid4( ))

        if address is not None:
            kwargs['backend_ip'] = address[0]
            kwargs['backend_port'] = address[1]

        super( ).__init__( *args, **kwargs )

        self.__send_queue = { }
        self.__pending = { }
        self.__incoming_callbacks = { }
        self.__websocket = None
        self.__lock = threading.Lock( )
        self.__session = None
        self.__abort = abort

        if self.__abort is not None and not callable(self.__abort):
                raise RuntimeError(f'abort function must be callable ({type(self.__abort)} is not)')

        # Fetch URL in Colab
        if is_colab( ):
            from google.colab.output import eval_js
            # Colab maps the internal port to a secure external URL
            external_https = eval_js(f"google.colab.kernel.proxyPort({self.backend_port})")
            print(f"Colab Proxy URL for {self.backend_port} is: {external_https}")
            self.backend_url = external_https.replace("https://", "wss://")
        else:
            # Standard local/remote access
            self.backend_url = f"ws://{self.backend_ip}:{self.backend_port}"

    def __enqueue_send( self, ident, msg, callback ):
        ### it is assumed that this is called AFTER the lock has been aquired
        if ident in self.__send_queue:
            self.__send_queue[ident].insert(0, { 'cb': callback, 'msg': msg })
        else:
            self.__send_queue[ident] = [ { 'cb': callback, 'msg': msg } ]
    def __dequeue_send( self, ident ):
        ### it is assumed that this is called AFTER the lock has been aquired
        if ident in self.__send_queue and self.__send_queue[ident]:
            return self.__send_queue[ident].pop( )
        return None
    async def __put_pending( self, ident, callback ):
        ### it is assumed that this is called AFTER the lock has been aquired
        ## info about request sent to javascript waits in this queue
        ## until the javascript reply is received...
        if ident in self.__pending:
            if self.__websocket is not None:
                await self.__websocket.send( { 'id': '', 'message': 'queueing callback, but already one callback waiting', 'direction': 'error' } )
        else:
            self.__pending[ident] = callback

    def __get_pending( self, ident ):
        ### it is assumed that this is called AFTER the lock has been aquired
        ## when reply arrives for a request to javascript
        ## the callback is retrieved from the waiting queue
        if ident in  self.__pending:
            result = self.__pending[ident]
            del self.__pending[ident]
            return result
        return None

    @classmethod
    async def __is_websocket_alive(cls, websocket):
        """Check if a websocket connection is still alive"""
        try:
            # Try to ping the websocket with a short timeout
            await asyncio.wait_for(websocket.ping(), timeout=2.0)
            return True
        except (asyncio.TimeoutError, ConnectionError, Exception):
            return False

    @classmethod
    def __cleanup_dead_sessions(cls):
        """Remove dead sessions from tracking"""
        current_time = time.time()
        dead_sessions = []

        for session_id, session_info in cls._active_sessions.items():
            # Remove sessions older than 5 minutes (stale cleanup)
            if current_time - session_info['timestamp'] > 300:
                dead_sessions.append(session_id)

        for session_id in dead_sessions:
            del cls._active_sessions[session_id]

    async def __handle_session_conflict(self, websocket, existing_session_info, new_session_id):
        """Handle session conflict by checking if existing connection is alive"""
        existing_ws = existing_session_info['websocket']
        existing_datapipe = existing_session_info['datapipe']

        # Check if existing websocket is still alive
        if await self.__is_websocket_alive(existing_ws):
            # Existing connection is alive - send conflict message to BOTH connections
            conflict_msg = {
                'id': 'session_conflict',
                'message': {
                    'type': 'session_conflict',
                    'error': 'Multiple windows/tabs detected. Please use only one browser window.',
                    'action': 'close_duplicate'
                },
                'direction': 'error'
            }

            try:
                # Send to existing connection
                await existing_ws.send(serialize(conflict_msg))

                # Send to new connection
                await websocket.send(serialize(conflict_msg))

                # Close the new connection
                await websocket.close(code=1008, reason='Session conflict')

                # Call abort on the existing DataPipe instance
                if existing_datapipe.__abort is not None:
                    err = RuntimeError(f"Session conflict detected: New connection attempted with session {new_session_id}")
                    existing_datapipe.__abort(err)

                return False  # Reject new connection

            except Exception as e:
                print(f"Error handling session conflict: {e}")
                # If we can't communicate, treat existing as dead
                pass

        # Existing connection is dead - replace it
        return True  # Allow new connection

    def register( self, ident, callback ):
        """Register a callback to handle all requests coming from JavaScript. The
        callback will be called whenever a request arrives.

        Parameters
        ----------
        ident: string
            UUID which is associated with the messages that should be delivered
            to this callback
        callback: (string) => dictionary
            Callback which receives the message from Python as its sole parameter.
            The return value of this callback is delivered to the JavaScript code
            as the ''reply'' to to JavaScript in response to the ''request'' contained
            in the message.
        """
        with self.__lock:
            self.__incoming_callbacks[ident] = callback

    async def send( self, ident, message, callback ):
        """Send a `message` to JavaScript identified by `ident`. Once the reply is
        received, the `callback` will be called with the reply message.

        Parameters
        ----------
        ident: string
            UUID to associate with this message. It is used to keep track of the callback
            that should be called when the reply is received.
        message: dictionary
            This dictionary contains the request for the JavaScript code.
        callback: (string) => void
            Callback which receives the message that JavaScript generates in response to
            the request contained in the `message` parameter.
        """
        with self.__lock:
            if self.__websocket is not None:
                msg = { 'id': ident, 'message': message, 'direction': 'p2j' }
                if ident in self.__pending:
                    self.__enqueue_send( ident, msg, callback )
                else:
                    if ident in self.__send_queue and self.__send_queue[ident]:
                        self.__enqueue_send( ident, msg, callback )
                        existing = self.__dequeue_send(ident)
                        self.__put_pending(ident, existing['cb'])
                        await self.__websocket.send(serialize( existing['msg'] ))
                    else:
                        await self.__put_pending(ident, callback)
                        await self.__websocket.send(serialize( msg ))

    async def process_messages( self, websocket ):
        """Process messages related to image display updates.

        Parameters
        ----------
        websocket: websocket object
            Websocket serve function passes in the websocket object to use.
        path:
            Websocket serve provides a second parameter.
        """
        try:
            self.__websocket = websocket
            session_established = False

            print(f"✓ WebSocket server received connection on port {self.backend_port}")

            async for message in websocket:
                msg = deserialize(message)
                if 'session' not in msg:
                    await self.__websocket.close( )
                    err = RuntimeError(f'session not in: {msg}')
                    if self.__abort is not None:
                        self.__abort( err )
                    else:
                        raise err
                    return

                # Handle session initialization with conflict detection
                if not session_established:
                    new_session_id = msg['session']

                    with self._session_lock:
                        # Clean up any stale sessions first
                        self.__cleanup_dead_sessions()

                        # Check if session already exists
                        if new_session_id in self._active_sessions:
                            existing_session_info = self._active_sessions[new_session_id]

                            # Handle the conflict
                            if not await self.__handle_session_conflict(websocket, existing_session_info, new_session_id):
                                return  # Connection was rejected

                        # Register this session as active
                        self._active_sessions[new_session_id] = {
                            'websocket': websocket,
                            'timestamp': time.time(),
                            'datapipe': self
                        }
                        self.__session = new_session_id
                        session_established = True

                # Existing session validation (your original logic)
                elif self.__session != None and self.__session != msg['session']:
                    conflict_msg = {
                        'id': self.__session,
                        'message': {
                            'type': 'session_corruption',
                            'error': 'Session corruption detected',
                            'expected': self.__session,
                            'received': msg['session']
                        },
                        'direction': 'error'
                    }

                    await self.__websocket.send(serialize(conflict_msg))
                    await self.__websocket.close()

                    # Clean up from active sessions
                    with self._session_lock:
                        if self.__session in self._active_sessions:
                            del self._active_sessions[self.__session]

                    err = RuntimeError(f"session corruption: {msg['session']} does not equal {self.__session}")
                    if self.__abort is not None:
                        self.__abort( err )
                    else:
                        raise err
                    return

                with self.__lock:
                    ###
                    ### initialize session identifier (pre-state-corruption fixes)
                    ###
                    if self.__session == None:
                        self.__session = msg['session']

                    if msg['direction'] == 'p2j':
                        cb = self.__get_pending(msg['id'])
                        outgo = self.__dequeue_send(msg['id'])
                        if outgo is not None:
                            await websocket.send(serialize(outgo['msg']))
                            await self.__put_pending(msg['id'],outgo['cb'])
                        if cb is not None:
                            if inspect.isawaitable(cb):
                                await cb(msg['message'])
                            else:
                                cb(msg['message'])
                    elif msg['id'] != 'initialize':
                        ###
                        ### no cb/reply for initialization of session
                        ###
                        if msg['id'] not in self.__incoming_callbacks:
                            raise RuntimeError(f'incoming js request with no callback: {msg}')
                        result = self.__incoming_callbacks[msg['id']](msg['message'])
                        if inspect.isawaitable(result):
                            try:
                                return_message = "an exception occurred in creating the respone"
                                return_message = await result
                                await self.__websocket.send( serialize( { 'id': msg['id'],
                                                                          'message': return_message,
                                                                          'direction': msg['direction'] } ) )
                            except Exception as e:
                                trace_back = traceback.format_exc().replace('\n','\n                       ')
                                print('************************************************************************************************************************')
                                print( f'''EXCEPTION ENCOUNTERED: {repr(e)}''' )
                                print( f'''              MESSAGE: {repr(msg)}''' )
                                print( f'''          STACK TRACE: {trace_back}''' )
                                print('************************************************************************************************************************')
                                await self.__websocket.send( serialize( { 'id': msg['id'],
                                                                          'message': { 'error': "exception encountered",
                                                                                       'errant': str(return_message),
                                                                                       'exception': repr(e) },
                                                                          'direction': str(msg['direction']) } ) )
                        else:
                            try:
                                await self.__websocket.send( serialize( { 'id': msg['id'],
                                                                          'message': result,
                                                                          'direction': msg['direction'] } ) )
                            except Exception as e:
                                trace_back = traceback.format_exc().replace('\n','\n                       ')
                                print('************************************************************************************************************************')
                                print( f'''EXCEPTION ENCOUNTERED: {repr(e)}''' )
                                print( f'''              MESSAGE: {repr(msg)}''' )
                                print( f'''          STACK TRACE: {trace_back}''' )
                                print('************************************************************************************************************************')
                                await self.__websocket.send( serialize( { 'id': msg['id'],
                                                                          'message': { 'error': "exception encountered",
                                                                                       'errant': str(result),
                                                                                       'exception': repr(e) },
                                                                          'direction': str(msg['direction']) } ) )
        finally:
            # Clean up when connection closes
            if self.__session is not None:
                with self._session_lock:
                    if self.__session in self._active_sessions:
                        del self._active_sessions[self.__session]
            self.__websocket = None
